package gov.va.shared.web;

import java.time.ZoneId;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.stereotype.Controller;
import org.springframework.ui.ModelMap;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;

import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.AppUser;
import gov.va.shared.model.AppUser.AppUserView;
import gov.va.shared.model.AppUserFacility;
import gov.va.shared.model.AppUserGlobalRole;
import gov.va.shared.model.Role;
import gov.va.shared.model.Role.RoleType;
import gov.va.shared.model.ldap.LdapPerson;
import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.shared.persistence.AppUserDAO;
import gov.va.shared.persistence.AppUserFacilityDAO;
import gov.va.shared.persistence.GrantableRoleDAO;
import gov.va.shared.persistence.LdapPersonDAO;
import gov.va.shared.persistence.lookup.RoleDAO;
import gov.va.shared.persistence.lookup.sds.VAFacilityDAO;
import gov.va.shared.service.AppUserService;
import gov.va.shared.service.CreateOrRetrieveUserCallback;
import gov.va.shared.util.PersistenceUtil;
import gov.va.shared.util.SecurityUtil;
import gov.va.shared.util.StringUtil;
import gov.va.shared.util.TimeZoneUtils;

@Controller
public class UserAdminController {

	@Autowired
	private AppUserDAO appUserDAO;
	@Autowired
	private AppUserFacilityDAO appUserFacilityDAO;
	@Autowired
	private AppUserService appUserService;
	@Autowired
	private GrantableRoleDAO grantableRoleDAO;
	@Autowired
	private LdapPersonDAO ldapPersonDAO;
	@Autowired
	private RoleDAO roleDAO;
	@Autowired
	private VAFacilityDAO vaFacilityDAO;

	/* Framework injection, if desired - CPB */
	@Autowired(required = false)
	private CreateOrRetrieveUserCallback createOrRetrieveUserCallback;

	@RequestMapping(value = "/userAdmin.htm", method = RequestMethod.GET)
	public String userAdmin(ModelMap modelMap) {
		modelMap.put("allTimeZones", TimeZoneUtils.TIME_ZONES);
		return "userAdmin";
	}

	@RequestMapping(value = "/appUser/{userId}", method = RequestMethod.GET)
	public @ResponseBody AppUser getAppUser(@PathVariable("userId") long userId) {
		ensureUserAccess(userId);

		return appUserDAO.findRequiredByPrimaryKey(userId);
	}

	private boolean hasUMPermission() {
		return SecurityUtil.getCurrentUser().hasGlobalRole(RoleType.USER_MANAGER);
	}

	@RequestMapping(value = "/appUserInfo", params = "updateRolesAndFacilities=true", method = RequestMethod.GET)
	@JsonView(AppUserView.Extended.class)
	public @ResponseBody Map<String, Object> getExtendedAppUserInfo(@RequestParam long userId) {
		return getAppUserInfo(userId, true);
	}

	@RequestMapping(value = "/appUserInfo", params = "updateRolesAndFacilities=false", method = RequestMethod.GET)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(AppUserView.Basic.class)
	public @ResponseBody Map<String, Object> getBasicAppUserInfo(@RequestParam long userId) {
		return getAppUserInfo(userId, false);
	}

	private Map<String, Object> getAppUserInfo(long userId, boolean updateRolesAndFacilities) {
		ensureUserAccess(userId);

		Map<String, Object> results = new HashMap<>();
		AppUser user = appUserDAO.findRequiredByPrimaryKey(userId);
		results.put("user", user);
		results.put("updateRolesAndFacilities", updateRolesAndFacilities);

		VAFacility primaryFacility = null;

		if (updateRolesAndFacilities) {
			SortedSet<VAFacility> availableFacilities = new TreeSet<>();
			availableFacilities.addAll(vaFacilityDAO.findAllSorted());

			List<AppUserFacility> appUserFacilityList = appUserFacilityDAO.findByUserSorted(user.getId());
			for (AppUserFacility facility : appUserFacilityList) {
				if (facility.isPrimaryFacility())
					primaryFacility = facility.getFacility();
				availableFacilities.remove(facility.getFacility());
			}

			results.put("availableFacilities", availableFacilities);
			results.put("appUserFacilities", appUserFacilityList);

			Set<Role> allGrantableRoles = new HashSet<>();

			SortedSet<Role> availableRoles = roleDAO.findAllSorted(true);
			Set<AppUserGlobalRole> globalRoles = user.getGlobalRoles();
			for (AppUserGlobalRole augr : globalRoles) {
				Role role = augr.getRole();
				availableRoles.remove(role);
				allGrantableRoles.addAll(grantableRoleDAO.findGrantableRolesForRole(role.getId()));
			}

			results.put("availableRoles", availableRoles);
			results.put("grantableRoles", allGrantableRoles);

			populateModelForSummaryTable(user, results);
		} else {
			primaryFacility = appUserFacilityDAO.findPrimaryFacilityForUser(user.getId());
		}
		results.put("defaultFacility", primaryFacility);

		return results;
	}

	private void populateModelForSummaryTable(AppUser user, Map<String, Object> results) {
		SortedSet<Role> allRoles = new TreeSet<>();
		allRoles.addAll(user.getBasicGlobalRoles());
		for (AppUserFacility f : user.getFacilities())
			allRoles.addAll(f.getRoles());

		Map<Long, Boolean> falseRoleMap = new HashMap<>();
		Map<Long, Role> roleInfoMap = new LinkedHashMap<>();
		for (Role r : allRoles) {
			falseRoleMap.put(r.getId(), false);
			roleInfoMap.put(r.getId(), r);
		}

		List<StationAndRoles> stationAndRoles = new ArrayList<>();

		for (VAFacility f : user.getBasicFacilities()) {
			Long vaFacilityId = f.getId();
			SortedSet<Role> rolesForFacility = user.getRolesForFacility(vaFacilityId);
			Map<Long, Boolean> roleMap = new HashMap<>(falseRoleMap);

			for (Role r : rolesForFacility)
				roleMap.put(r.getId(), true);
			stationAndRoles.add(new StationAndRoles(f.getId(), roleMap));
		}
		results.put("stationAndRoles", stationAndRoles);
		results.put("roleInfoMap", roleInfoMap);
	}

	public static class StationAndRoles {
		public long facilityId;
		public Map<Long, Boolean> roleMap;

		public StationAndRoles(long facilityId, Map<Long, Boolean> roleMap) {
			this.facilityId = facilityId;
			this.roleMap = roleMap;
		}
	}

	@RequestMapping(value = "/appUserUpdate", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean processUserUpdate(@RequestParam long userId, @RequestParam boolean enabled,
			@RequestParam boolean expired, @RequestParam boolean locked, @RequestParam ZoneId timezone,
			@RequestParam(required = false, defaultValue = "") List<Long> globalRoles,
			@RequestParam(required = false, defaultValue = "") List<Long> vaFacilitiesToAdd,
			@RequestParam(required = false, defaultValue = "") List<Long> vaFacilitiesToRemove,
			@RequestParam Long defaultFacilityId, @RequestParam boolean updateRolesAndFacilities) {
		ensureUserAccess(userId);

		AppUser user = appUserDAO.findRequiredByPrimaryKey(userId);

		Set<Long> vaFacilities = null;
		if (updateRolesAndFacilities) {
			vaFacilities = new HashSet<>(PersistenceUtil.translateObjectsToIds(user.getBasicFacilities()));
			for (Long l : vaFacilitiesToRemove)
				vaFacilities.remove(l);
			for (Long l : vaFacilitiesToAdd)
				vaFacilities.add(l);
		}

		appUserService.updateUser(userId, enabled != user.isEnabled() ? enabled : null,
				locked != user.isLocked() ? locked : null, expired != user.isAccountExpired() ? expired : null,
				timezone, updateRolesAndFacilities, defaultFacilityId, globalRoles, vaFacilities);
		return true;
	}

	private void ensureUserAccess(long userId) {
		long myUserId = SecurityUtil.getCurrentUser().getId();
		if (!hasUMPermission() && userId != myUserId)
			throw new AccessDeniedException("The user with the specified ID is not available");
	}

	@RequestMapping(value = "/appUserCustomize", method = RequestMethod.POST)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean processUserCustomize(@RequestParam long userId,
			@RequestParam(required = false, defaultValue = "") List<Long> roles,
			@RequestParam(required = false, defaultValue = "") List<Long> vaFacilities) {
		if (!hasUMPermission())
			throw new AccessDeniedException("The user with the specified ID is not available");

		appUserService.customizeUser(userId, roles, vaFacilities);
		return true;
	}

	@RequestMapping("/findAppUsers")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody Collection<AppUser> findAppUsers(@RequestParam(required = false) String name,
			@RequestParam(required = false) String activeDirectoryName) {
		if (StringUtils.isBlank(name) && StringUtils.isBlank(activeDirectoryName))
			throw new IllegalArgumentException("Please specify at least one piece of search criteria");

		SortedSet<AppUser> results = new TreeSet<>();

		if (StringUtils.isNotBlank(activeDirectoryName)) {
			LdapPerson person = ldapPersonDAO.findBySAMAccountName(activeDirectoryName);
			if (person != null) {
				AppUser appUser = new AppUser(person);
				results.add(appUser);
			}
		} else if (StringUtils.isNotBlank(name)) {
			String[] nameComponents = StringUtil.parseNameComponents(name);
			String lastName = nameComponents[0];
			String firstName = nameComponents[1];
			List<LdapPerson> ldapResults = ldapPersonDAO.findByName(firstName, lastName, false);
			for (LdapPerson p : ldapResults)
				results.add(new AppUser(p));
		}

		return results;
	}

	@RequestMapping("/appUserAdd")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody AppUser addAppUser(@RequestParam String activeDirectoryName) {
		if (StringUtils.isBlank(activeDirectoryName))
			throw new IllegalArgumentException("Active directory name is required");

		AppUser appUser = appUserService.createOrRetrieveUser(activeDirectoryName, createOrRetrieveUserCallback);
		return appUser;
	}

	@RequestMapping("/appUserRemove")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean removeAppUser(@RequestParam long appUserId) {
		boolean successful = appUserService.removeUser(appUserId);
		return successful;
	}

	@RequestMapping("/appUser")
	@JsonView(AppUserView.Basic.class)
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody SortedSet<AppUser> getAllAppUsers() {
		return appUserDAO.findAllSorted();
	}

}
