package gov.va.vss.model;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.time.LocalDate;

import javax.persistence.AttributeOverride;
import javax.persistence.AttributeOverrides;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.Table;
import javax.validation.constraints.NotNull;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.springframework.format.annotation.DateTimeFormat;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.vss.model.Volunteer.VolunteerView;
import gov.va.vss.model.lookup.VtkInstitution;
import gov.va.vss.util.DateUtil;

@Entity
@Table(name = "VtkVolunteerStationProfiles", schema = "dbo")
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
@AttributeOverrides({ @AttributeOverride(name = "createdDate", column = @Column(name = "EntryDate") ) })
public class VolunteerStationProfile extends AbstractAuditedVersionedPersistent<VolunteerStationProfile> {
	private static final long serialVersionUID = 6904844123870655771L;

	// -------------------------------------- Fields

	@NotNull
	private VtkInstitution vtkInstitution;
	@NotNull
	private Volunteer volunteer;
	@NotNull
	private VolunteerContactInfo contactInfo;
	private NextOfKin nextOfKin;
	private int yearsAtStation;
	private Integer hoursServedPriorYears;
	private Integer hoursServedCurrentYear;
	private Integer lastAwardHours;
	@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY)
	private LocalDate lastAwardDate;
	private Award award;
	@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY)
	private LocalDate terminationDate;
	private Boolean terminatedWithCause;
	private String remarks;
	private String transportationMethod;
	private Integer mealsEligible;
	private Integer hoursServedTotal;
	private Event event;

	public static class VolunteerStationProfileView {
		public static class Basic {
		}

		public static class Extended extends Basic {
		}

		public static class Search extends Basic {
		}
	}

	// -------------------------------------- Constructors

	public VolunteerStationProfile() {
	}

	/**
	 * Constructor for new VolunteerStationProfile - populates parent
	 * reference(s) and creates new empty children
	 * 
	 * @param volunteer
	 * @param vaFacility
	 */
	public VolunteerStationProfile(Volunteer volunteer, VtkInstitution vtkInstitution) {
		this.volunteer = volunteer;
		this.vtkInstitution = vtkInstitution;
		this.contactInfo = new VolunteerContactInfo(volunteer);
		this.nextOfKin = new NextOfKin(volunteer);
	}

	// -------------------------------------- Business Methods

	// -------------------------------------- Common Methods

	@Override
	protected boolean requiredEquals(VolunteerStationProfile oo) {
		return new EqualsBuilder().append(nullSafeGetId(vtkInstitution), nullSafeGetId(oo.getVtkInstitution()))
				.append(nullSafeGetId(volunteer), nullSafeGetId(oo.getVolunteer())).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(nullSafeGetId(getVtkInstitution())).append(nullSafeGetId(volunteer))
				.toHashCode();
	}

	// -------------------------------------- Accessor Methods

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "StationFK", nullable = false)
	@JsonView({ VolunteerStationProfileView.Extended.class, VolunteerStationProfileView.Search.class,
			VolunteerView.Search.class })
	public VtkInstitution getVtkInstitution() {
		return vtkInstitution;
	}

	public void setVtkInstitution(VtkInstitution vtkInstitution) {
		this.vtkInstitution = vtkInstitution;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "VolunteerFK", nullable = false)
	@JsonIgnore
	public Volunteer getVolunteer() {
		return volunteer;
	}

	public void setVolunteer(Volunteer volunteer) {
		this.volunteer = volunteer;
	}

	@ManyToOne(fetch = FetchType.LAZY, cascade = CascadeType.ALL)
	@JoinColumn(name = "VolunteerContactInfoFK", nullable = false, unique = true)
	@JsonView({ VolunteerStationProfileView.Search.class, VolunteerView.Search.class })
	public VolunteerContactInfo getContactInfo() {
		return contactInfo;
	}

	public void setContactInfo(VolunteerContactInfo contactInfo) {
		this.contactInfo = contactInfo;
	}

	@ManyToOne(fetch = FetchType.LAZY, cascade = CascadeType.ALL)
	@JoinColumn(name = "NextOfKinFK")
	@JsonView(VolunteerStationProfileView.Extended.class)
	public NextOfKin getNextOfKin() {
		return nextOfKin;
	}

	public void setNextOfKin(NextOfKin nextOfKin) {
		this.nextOfKin = nextOfKin;
	}

	public int getYearsAtStation() {
		return yearsAtStation;
	}

	public void setYearsAtStation(int yearsAtStation) {
		this.yearsAtStation = yearsAtStation;
	}

	@Column(name = "PriorYearsHoursServed")
	public Integer getHoursServedPriorYears() {
		return hoursServedPriorYears;
	}

	public void setHoursServedPriorYears(Integer hoursServedPriorYears) {
		this.hoursServedPriorYears = hoursServedPriorYears;
	}

	@Column(name = "CurrentYearHoursServed")
	public Integer getHoursServedCurrentYear() {
		return hoursServedCurrentYear;
	}

	public void setHoursServedCurrentYear(Integer hoursServedCurrentYear) {
		this.hoursServedCurrentYear = hoursServedCurrentYear;
	}

	@Column(name = "HoursLastAward")
	public Integer getLastAwardHours() {
		return lastAwardHours;
	}

	public void setLastAwardHours(Integer lastAwardHours) {
		this.lastAwardHours = lastAwardHours;
	}

	@Column(name = "DateLastAward")
	public LocalDate getLastAwardDate() {
		return lastAwardDate;
	}

	public void setLastAwardDate(LocalDate lastAwardDate) {
		this.lastAwardDate = lastAwardDate;
	}

	@ManyToOne(fetch = FetchType.LAZY, cascade = CascadeType.ALL)
	@JoinColumn(name = "AwardCodesFK")
	@JsonView(VolunteerStationProfileView.Extended.class)
	public Award getAward() {
		return award;
	}

	public void setAward(Award award) {
		this.award = award;
	}

	public LocalDate getTerminationDate() {
		return terminationDate;
	}

	public void setTerminationDate(LocalDate terminationDate) {
		this.terminationDate = terminationDate;
	}

	@Column(name = "IsTerminatedWithCause")
	public Boolean getTerminatedWithCause() {
		return terminatedWithCause;
	}

	public void setTerminatedWithCause(Boolean terminatedWithCause) {
		this.terminatedWithCause = terminatedWithCause;
	}

	@Column(length = 800)
	public String getRemarks() {
		return remarks;
	}

	public void setRemarks(String remarks) {
		this.remarks = remarks;
	}

	@Column(name = "MethodOfTransportation", length = 1)
	public String getTransportationMethod() {
		return transportationMethod;
	}

	public void setTransportationMethod(String transportationMethod) {
		this.transportationMethod = transportationMethod;
	}

	@Column(name = "EligibleForMeals")
	public Integer getMealsEligible() {
		return mealsEligible;
	}

	public void setMealsEligible(Integer mealsEligible) {
		this.mealsEligible = mealsEligible;
	}

	@Column(name = "TotalHoursServed")
	public Integer getHoursServedTotal() {
		return hoursServedTotal;
	}

	public void setHoursServedTotal(Integer hoursServedTotal) {
		this.hoursServedTotal = hoursServedTotal;
	}

	@ManyToOne(fetch = FetchType.LAZY, cascade = CascadeType.ALL)
	@JoinColumn(name = "EventsFK")
	@JsonView(VolunteerStationProfileView.Extended.class)
	public Event getEvent() {
		return event;
	}

	public void setEvent(Event event) {
		this.event = event;
	}

}
