package gov.va.vss.config;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.annotation.Resource;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.MessageSource;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.security.authentication.AccountExpiredException;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.authentication.CredentialsExpiredException;
import org.springframework.security.authentication.DisabledException;
import org.springframework.security.authentication.LockedException;
import org.springframework.security.authentication.event.LoggerListener;
import org.springframework.security.config.annotation.authentication.builders.AuthenticationManagerBuilder;
import org.springframework.security.config.annotation.web.builders.HttpSecurity;
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity;
import org.springframework.security.core.AuthenticationException;
import org.springframework.security.core.userdetails.AuthenticationUserDetailsService;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.core.userdetails.UsernameNotFoundException;
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.security.web.access.AccessDeniedHandler;
import org.springframework.security.web.access.AccessDeniedHandlerImpl;
import org.springframework.security.web.authentication.AuthenticationFailureHandler;
import org.springframework.security.web.authentication.ExceptionMappingAuthenticationFailureHandler;
import org.springframework.security.web.authentication.LoginUrlAuthenticationEntryPoint;
import org.springframework.security.web.authentication.preauth.PreAuthenticatedAuthenticationProvider;
import org.springframework.security.web.authentication.preauth.PreAuthenticatedAuthenticationToken;
import org.springframework.security.web.authentication.preauth.RequestHeaderAuthenticationFilter;

import gov.va.shared.config.AbstractSecurityConfig;
import gov.va.shared.config.HybridDaoAuthenticationProvider;
import gov.va.shared.model.Permission;
import gov.va.shared.web.DatabaseDrivenPreAuthenticationFilter;
import gov.va.vss.web.ajax.AjaxRequestHandler;

@Configuration
@EnableWebSecurity
public class WebSecurityConfig extends AbstractSecurityConfig {

	public static final String URI_LOGIN = "/login.htm";
	public static final String URI_LOGOUT = "/logout.htm";
	public static final String URI_DEFAULT = "/index.htm";
	public static final String URI_AUTH_EXCEPTION = "/processAuthorizationException.htm";
	public static final String MEDIA_DIR = "/media";

	@Value("${port.http}")
	private int httpPort;
	@Value("${port.https}")
	private int httpsPort;
	@Value("${httpsRequired}")
	private boolean httpsRequired;
	@Value("${cookie.sessionId")
	private String cookieSessionId;
	@Resource(name = "messageSource")
	private MessageSource source;

	@Autowired
	private PasswordEncoder passwordEncoder;

	public static void main(String[] args) throws Exception {
		System.out.println(new BCryptPasswordEncoder().encode("Primer"));
	}

	@Bean
	public AccessDeniedHandler myAccessDeniedHandler() {
		AccessDeniedHandlerImpl adh = new AccessDeniedHandlerImpl();
		adh.setErrorPage(URI_AUTH_EXCEPTION);
		return adh;
	}

	@Bean
	public AuthenticationFailureHandler myAuthenticationFailureHandler() {
		ExceptionMappingAuthenticationFailureHandler h = new ExceptionMappingAuthenticationFailureHandler();

		Map<String, String> failureUrlMap = new HashMap<>();
		failureUrlMap.put(AccountExpiredException.class.getName(), URI_LOGIN + "?error=expired");
		failureUrlMap.put(BadCredentialsException.class.getName(), URI_LOGIN + "?error=default");
		failureUrlMap.put(CredentialsExpiredException.class.getName(), URI_LOGIN + "?error=expired");
		failureUrlMap.put(DisabledException.class.getName(), URI_LOGIN + "?error=disabled");
		failureUrlMap.put(LockedException.class.getName(), URI_LOGIN + "?error=locked");
		h.setExceptionMappings(failureUrlMap);

		h.setDefaultFailureUrl(URI_LOGIN + "?error=default");
		return h;
	}

	@Override
	protected void configure(HttpSecurity http) throws Exception {
		http.authorizeRequests() //
				.antMatchers(URI_LOGIN, URI_LOGOUT, URI_AUTH_EXCEPTION, MEDIA_DIR + "/**").permitAll() //
				.antMatchers(URI_DEFAULT, "/selectStation.htm", WebConfig.AJAX_CONTEXT_PATH_PREFIX + "/**")
				.authenticated() //
				.antMatchers("/**/*.htm").hasAuthority(Permission.LOGIN_APPLICATION) //
				.and().formLogin().loginPage(URI_LOGIN).defaultSuccessUrl(URI_DEFAULT, true)
				.failureHandler(myAuthenticationFailureHandler());

		http.addFilterBefore(preauthFilter(), RequestHeaderAuthenticationFilter.class);

		http.logout().invalidateHttpSession(true).deleteCookies(cookieSessionId);

		http.exceptionHandling().authenticationEntryPoint(new AjaxAwareAuthenticationEntryPoint(URI_LOGIN)) //
				.accessDeniedHandler(myAccessDeniedHandler());
		
		if (httpsRequired) {
			/*
			 * Use HTTPs for ALL requests
			 */
			http.requiresChannel().anyRequest().requiresSecure();
			http.portMapper().http(httpPort).mapsTo(httpsPort);
		}
	}

	@Bean
	public DatabaseDrivenPreAuthenticationFilter preauthFilter() throws Exception {
		DatabaseDrivenPreAuthenticationFilter f = new DatabaseDrivenPreAuthenticationFilter(MEDIA_DIR);
		f.setAuthenticationManager(authenticationManagerBean());
		return f;
	}

	@Bean
	public LoggerListener loggerListener() {
		return new LoggerListener();
	}

	@Override
	protected void configure(AuthenticationManagerBuilder auth) throws Exception {
		auth.eraseCredentials(true) //
				.authenticationProvider(preAuthenticatedProvider()) //
				.authenticationProvider(hybridAuthenticationProvider());
	}

	@Bean
	public PreAuthenticatedAuthenticationProvider preAuthenticatedProvider() {
		PreAuthenticatedAuthenticationProvider p = new PreAuthenticatedAuthenticationProvider();
		p.setPreAuthenticatedUserDetailsService(
				new AuthenticationUserDetailsService<PreAuthenticatedAuthenticationToken>() {
					@Override
					public UserDetails loadUserDetails(PreAuthenticatedAuthenticationToken token)
							throws UsernameNotFoundException {
						return (UserDetails) token.getPrincipal();
					}
				});
		return p;
	}

	@Bean
	public HybridDaoAuthenticationProvider hybridAuthenticationProvider() {
		HybridDaoAuthenticationProvider p = new HybridDaoAuthenticationProvider(passwordEncoder);
		return p;
	}

	public static class AjaxAwareAuthenticationEntryPoint extends LoginUrlAuthenticationEntryPoint {
		public AjaxAwareAuthenticationEntryPoint(String loginUrl) {
			super(loginUrl);
		}

		@Override
		public void commence(HttpServletRequest request, HttpServletResponse response,
				AuthenticationException authException) throws IOException, ServletException {
			if (AjaxRequestHandler.isAjax(request)) {
				response.sendError(HttpServletResponse.SC_FORBIDDEN, "Request Denied (Session Expired)");
			} else {
				super.commence(request, response, authException);
			}
		}
	}

}
