package gov.va.vss.model;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.time.LocalDate;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;

import org.apache.commons.lang3.builder.CompareToBuilder;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.annotations.BatchSize;
import org.hibernate.validator.constraints.Length;
import org.hibernate.validator.constraints.NotBlank;
import org.springframework.format.annotation.DateTimeFormat;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.shared.util.StringUtil;
import gov.va.vss.model.lookup.VtkLanguage;
import gov.va.vss.util.DateUtil;

@Entity
@Table(name = "VtkVolunteers", schema = "dbo")
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public class Volunteer extends AbstractAuditedVersionedPersistent<Volunteer>
		implements Comparable<Volunteer>, ObjectScopedToStationNumbers {
	private static final long serialVersionUID = 6904844123870655771L;

	// -------------------------------------- Fields

	@NotBlank
	private String lastName;
	@NotBlank
	private String firstName;
	private String middleName;
	private String suffix;
	@NotBlank
	private String sexMF;
	@NotNull
	@DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY)
	private LocalDate dateOfBirth;
	private String dateOfBirthImprecision;
	private String nickname;
	private VtkLanguage preferredLanguage;
	private Integer d0;
	private String remarks;
	private Boolean vaEmployee;
	private String displayNameLowercase;
	private String displayNameUppercase;
	private Integer age;
	@Length(max = 5)
	private String identifyingCode;
	private String fullNameForReports;

	private List<VolunteerStationProfile> stationProfiles;

	// -------------------------------------- Business Methods

	@Transient
	public Collection<String> getScopedToStationNumbers() {
		List<String> results = new ArrayList<>();
		for (VolunteerStationProfile profile : getStationProfiles()) {
			results.add(profile.getVtkInstitution().getStationNumber());
		}
		return results;
	}

	@Transient
	public String getDisplayName() {
		return StringUtil.getDisplayName(true, firstName, middleName, lastName, suffix);
	}

	@Transient
	public VolunteerStationProfile getStationProfileForSta3n(String sta3n) {
		for (VolunteerStationProfile profile : getStationProfiles()) {
			VAFacility vaFacility = profile.getVtkInstitution().getVaFacility();
			if (vaFacility != null && sta3n.equals(vaFacility.getSta3n()))
				return profile;
		}
		return null;
	}

	/**
	 * See @JsonView documentation - CPB
	 */
	public static class VolunteerView {
		public static class Basic {
		}

		public static class Extended extends Basic {
		}

		public static class Search extends Basic {
		}
	}

	// -------------------------------------- Common Methods

	@Override
	protected boolean requiredEquals(Volunteer oo) {
		return new EqualsBuilder().append(getDisplayName(), oo.getDisplayName())
				.append(getDateOfBirth(), oo.getDateOfBirth()).append(getIdentifyingCode(), oo.getIdentifyingCode())
				.isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(getIdentifyingCode()).toHashCode();
	}

	@Override
	public int compareTo(Volunteer o) {
		if (equals(o))
			return 0;

		return new CompareToBuilder().append(nullSafeLowercase(getDisplayName()), nullSafeLowercase(o.getDisplayName()))
				.toComparison() > 0 ? 1 : -1;
	}

	public String toString() {
		return getDisplayName() + " (DOB " + getDateOfBirth() + ", code '" + identifyingCode + "')";
	}

	// -------------------------------------- Accessor Methods

	@Column(length = 30)
	public String getLastName() {
		return lastName;
	}

	public void setLastName(String lastName) {
		this.lastName = lastName;
	}

	@Column(length = 30)
	public String getFirstName() {
		return firstName;
	}

	public void setFirstName(String firstName) {
		this.firstName = firstName;
	}

	@Column(length = 20)
	public String getMiddleName() {
		return middleName;
	}

	public void setMiddleName(String middleName) {
		this.middleName = middleName;
	}

	@Column(name = "NameSuffix", length = 10)
	public String getSuffix() {
		return suffix;
	}

	public void setSuffix(String suffix) {
		this.suffix = suffix;
	}

	@Column(name = "Sex", length = 1)
	public String getSexMF() {
		return sexMF;
	}

	public void setSexMF(String sexMF) {
		this.sexMF = sexMF;
	}

	public LocalDate getDateOfBirth() {
		return dateOfBirth;
	}

	public void setDateOfBirth(LocalDate dateOfBirth) {
		this.dateOfBirth = dateOfBirth;
	}

	public String getDateOfBirthImprecision() {
		return dateOfBirthImprecision;
	}

	public void setDateOfBirthImprecision(String dateOfBirthImprecision) {
		this.dateOfBirthImprecision = dateOfBirthImprecision;
	}

	@Column(name = "    Name", length = 20)
	public String get    name() {
		return nickname;
	}

	public void set    name(String nickname) {
		this.nickname = nickname;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "PreferredLanguageForLoginFK")
	@JsonView(VolunteerView.Extended.class)
	public VtkLanguage getPreferredLanguage() {
		return preferredLanguage;
	}

	public void setPreferredLanguage(VtkLanguage preferredLanguage) {
		this.preferredLanguage = preferredLanguage;
	}

	public Integer getD0() {
		return d0;
	}

	public void setD0(Integer d0) {
		this.d0 = d0;
	}

	@Column(length = 1000)
	public String getRemarks() {
		return remarks;
	}

	public void setRemarks(String remarks) {
		this.remarks = remarks;
	}

	@Column(name = "IsVaEmployee")
	public Boolean getVaEmployee() {
		return vaEmployee;
	}

	public void setVaEmployee(Boolean vaEmployee) {
		this.vaEmployee = vaEmployee;
	}

	@Column(name = "LowercaseNameForDisplay", length = 120)
	public String getDisplayNameLowercase() {
		return displayNameLowercase;
	}

	public void setDisplayNameLowercase(String displayNameLowercase) {
		this.displayNameLowercase = displayNameLowercase;
	}

	@Column(name = "UppercaseNameForDisplay", length = 120)
	public String getDisplayNameUppercase() {
		return displayNameUppercase;
	}

	public void setDisplayNameUppercase(String displayNameUppercase) {
		this.displayNameUppercase = displayNameUppercase;
	}

	@Column(insertable = false, updatable = false)
	public Integer getAge() {
		return age;
	}

	public void setAge(Integer age) {
		this.age = age;
	}

	@Column(length = 5)
	public String getIdentifyingCode() {
		return identifyingCode;
	}

	public void setIdentifyingCode(String identifyingCode) {
		this.identifyingCode = identifyingCode;
	}

	@Column(length = 120)
	public String getFullNameForReports() {
		return fullNameForReports;
	}

	public void setFullNameForReports(String fullNameForReports) {
		this.fullNameForReports = fullNameForReports;
	}

	@OneToMany(fetch = FetchType.LAZY, mappedBy = "volunteer", cascade = CascadeType.ALL)
	@BatchSize(size = 500)
	@JsonView({ VolunteerView.Extended.class, VolunteerView.Search.class })
	public List<VolunteerStationProfile> getStationProfiles() {
		if (stationProfiles == null)
			stationProfiles = new ArrayList<>();
		return stationProfiles;
	}

	public void setStationProfiles(List<VolunteerStationProfile> stationProfiles) {
		this.stationProfiles = stationProfiles;
	}

}
