package gov.va.vss.model;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Convert;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Pattern;

import org.apache.commons.lang3.builder.CompareToBuilder;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.annotations.BatchSize;
import org.hibernate.validator.constraints.Email;
import org.hibernate.validator.constraints.Length;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.core.AbstractAuditedVersionedPersistent;
import gov.va.shared.util.StringUtil;
import gov.va.vss.model.lookup.VtkDonorType;
import gov.va.vss.model.lookup.VtkDonorType.VtkDonorTypeValue;
import gov.va.vss.persistence.conversion.ScopeTypeConverter;
import gov.va.vss.util.ValidationUtil;
import gov.va.vss.web.conversion.VtkDonorTypeConverter;
import gov.va.vss.model.lookup.VtkState;

@Entity
@Table(name = "VssDonor", schema = "dbo")
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public class Donor extends AbstractAuditedVersionedPersistent<Donor> implements Comparable<Donor> {
	private static final long serialVersionUID = 583796042812902141L;

	/**
	 * See @JsonView documentation - CPB
	 */
	public static class DonorView {
		public interface Basic {
		}

		public interface Extended extends Basic {
		}
	}

	// -------------------------------------- Fields

	@Length(max = 30)
	private String lastName;
	@Length(max = 30)
	private String firstName;
	@Length(max = 20)
	private String middleName;
	private String prefix;
	@Length(max = 10)
	private String suffix;
	@Length(max = 50)
	private String donorAffiliation;
	@Length(max = 35)
	private String addressLine1;
	@Length(max = 35)
	private String addressLine2;
	@Length(max = 30)
	private String city;
	@Length(max = 10)
	private String zip;
	@Length(max = 250)
	@Email(message = "Please enter a valid email in the format 'user@domain.tld'.")
	private String email;
	@Length(max = 30)
	@Pattern(regexp = ValidationUtil.PHONE_REGEX, message = "Please enter a valid phone number.")
	private String phone;

	private VtkDonorType vtkDonorType;
	private Organization organization;
	private VtkState vtkState;
	private Volunteer volunteer;

	private List<DonationSummary> donations;

	// -------------------------------------- Business Methods

	@Transient
	public String getAddressMultilineDisplay() {
		return getAddressDisplay(false);
	}

	private String getAddressDisplay(boolean useIdForState) {
		return StringUtil.getAddressDisplay(addressLine1, addressLine2, null, city,
				useIdForState ? String.valueOf(nullSafeGetId(getVtkState())) : (getVtkState()!=null ? getVtkState().getAbbreviation() : ""), zip,
				"\n");
	}

	@Transient
	@JsonView(DonorView.Basic.class)
	public String getDisplayName() {
		VtkDonorTypeValue t = getVtkDonorType().getLookupType();
		if (t == VtkDonorTypeValue.ORGANIZATION)
			return getOrganization().getDisplayName();

		Volunteer v = getVolunteer();
		if (v != null)
			return v.getDisplayName();

		return StringUtil.getDisplayName(true, firstName, middleName, lastName, suffix);
	}

	@Transient
	@JsonIgnore
	private boolean isDemographicsUsed() {
		return getVtkDonorType().getLookupType() == VtkDonorTypeValue.INDIVIDUAL && getVolunteer() == null;
	}

	// -------------------------------------- Common Methods

	@Override
	public int compareTo(Donor o) {
		if (equals(o))
			return 0;

		return new CompareToBuilder().append(nullSafeLowercase(getDisplayName()), nullSafeLowercase(o.getDisplayName()))
				.toComparison() > 0 ? 1 : -1;
	}

	@Override
	protected boolean requiredEquals(Donor oo) {
		String displayName = isDemographicsUsed() ? nullSafeLowercase(getDisplayName()) : "";
		String otherDisplayName = oo.isDemographicsUsed() ? nullSafeLowercase(oo.getDisplayName()) : "";
		String addressStr = isDemographicsUsed() ? nullSafeLowercase(getAddressDisplay(true)) : "";
		String otherAddressStr = oo.isDemographicsUsed() ? nullSafeLowercase(oo.getAddressDisplay(true)) : "";

		return new EqualsBuilder().append(nullSafeGetId(getVtkDonorType()), nullSafeGetId(oo.getVtkDonorType()))
				.append(nullSafeGetId(getVolunteer()), nullSafeGetId(oo.getVolunteer()))
				.append(nullSafeGetId(getOrganization()), nullSafeGetId(oo.getOrganization()))
				.append(displayName, otherDisplayName).append(addressStr, otherAddressStr).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		String displayName = isDemographicsUsed() ? nullSafeLowercase(getDisplayName()) : "";
		String addressStr = isDemographicsUsed() ? nullSafeLowercase(getAddressDisplay(true)) : "";

		return new HashCodeBuilder().append(nullSafeGetId(getVtkDonorType())).append(nullSafeGetId(getVolunteer()))
				.append(nullSafeGetId(getOrganization())).append(displayName).append(addressStr).toHashCode();
	}

	public String toString() {
		return getDisplayName();
	}

	// -------------------------------------- Accessor Methods

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "VssVolunteersFK", unique = true, updatable = false, insertable = false)

	@JsonIgnore
	public Volunteer getVolunteer() {
		return volunteer;
	}

	public void setVolunteer(Volunteer volunteer) {
		this.volunteer = volunteer;
	}

	@Convert(converter = VtkDonorTypeConverter.class)
	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "DonorTypeFK", unique = true)
	@JsonIgnore
	@NotNull
	public VtkDonorType getVtkDonorType() {
		return vtkDonorType;
	}

	public void setVtkDonorType(VtkDonorType vtkDonorType) {
		this.vtkDonorType = vtkDonorType;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "OrganizationFK", unique = true, updatable = false, insertable = false)
	@JsonIgnore
	public Organization getOrganization() {
		return organization;
	}

	public void setOrganization(Organization organization) {
		this.organization = organization;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "STD_StateFK", unique = true)
	@JsonIgnore
	public VtkState getVtkState() {
		return vtkState;
	}

	public void setVtkState(VtkState vtkState) {
		this.vtkState = vtkState;
	}

	@Column(name = "NamePrefix", length = 10)
	public String getPrefix() {
		return prefix;
	}

	public void setPrefix(String prefix) {
		this.prefix = prefix;
	}

	@Column(name = "NameSuffix", length = 10)
	public String getSuffix() {
		return suffix;
	}

	public void setSuffix(String suffix) {
		this.suffix = suffix;
	}

	// __JC_TODO__ ?!!! what??
	@Column(name = "DonorAffiliation", length = 50)
	public String getDonorAffiliation() {
		return donorAffiliation;
	}

	public void setDonorAffiliation(String donorAffiliation) {
		this.donorAffiliation = donorAffiliation;
	}

	@Column(name = "Address1", length = 35)
	public String getAddressLine1() {
		return addressLine1;
	}

	public void setAddressLine1(String address1) {
		this.addressLine1 = address1;
	}

	@Column(name = "Address2", length = 35)
	public String getAddressLine2() {
		return addressLine2;
	}

	public void setAddressLine2(String address2) {
		this.addressLine2 = address2;
	}

	@Column(name = "City", length = 30)
	public String getCity() {
		return city;
	}

	public void setCity(String city) {
		this.city = city;
	}

	@Column(name = "Zip", length = 10)
	public String getZip() {
		return zip;
	}

	public void setZip(String zip) {
		this.zip = zip;
	}

	@Column(name = "Email", length = 250)
	public String getEmail() {
		return email;
	}

	public void setEmail(String email) {
		this.email = email;
	}

	@Column(name = "Phone", length = 30)
	public String getPhone() {
		return phone;
	}

	public void setPhone(String phone) {
		this.phone = phone;
	}

	@Column(name = "LastName", length = 30)
	public String getLastName() {
		return lastName;
	}

	public void setLastName(String lastName) {
		this.lastName = lastName;
	}

	@Column(name = "firstName", length = 30)
	public String getFirstName() {
		return firstName;
	}

	public void setFirstName(String firstName) {
		this.firstName = firstName;
	}

	@Column(name = "MiddleName", length = 20)
	public String getMiddleName() {
		return middleName;
	}

	public void setMiddleName(String middleName) {
		this.middleName = middleName;
	}

	@OneToMany(fetch = FetchType.LAZY, mappedBy = "donor", cascade = CascadeType.ALL)
	@BatchSize(size = 100)
	public List<DonationSummary> getDonations() {
		if (donations == null)
			donations = new ArrayList<>();
		return donations;
	}

	public void setDonations(List<DonationSummary> donations) {
		this.donations = donations;
	}
}
