package gov.va.vss.model.lookup;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.annotations.BatchSize;
import org.hibernate.annotations.Cache;
import org.hibernate.annotations.CacheConcurrencyStrategy;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;

import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.shared.util.StationsUtil;
import gov.va.vss.model.DonGenPostFund;
import gov.va.vss.model.DonationReference;
import gov.va.vss.model.StationParameters;
import gov.va.vss.model.VoluntaryServiceParameters;

@Entity
@Table(name = "VtkInstitutions_tbl", schema = "dbo")
@Cache(usage = CacheConcurrencyStrategy.NONSTRICT_READ_WRITE)
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public class VtkInstitution extends AbstractBasicInstitution<VtkInstitution> {
	private static final long serialVersionUID = -8678395783438462990L;

	/**
	 * Factory method to create a new VtkInstitution which reproduces all
	 * trigger-based logic in the DB (creating related entries) - CPB
	 * 
	 * @param facility
	 * @param state
	 * @param type
	 * @param parent
	 * @return
	 */
	public static VtkInstitution createNew(VAFacility facility, VtkState state, VtkInstitutionType type,
			VtkInstitution parent) {
		VtkInstitution newInstitution = new VtkInstitution(facility, state, type, null);
		newInstitution.populateMissingRequiredFields();
		return newInstitution;
	}

	public boolean populateMissingRequiredFields() {
		boolean dataChanged = false;

		StationParameters sp = getStationParameters();
		if (sp == null) {
			sp = new StationParameters(this);
			setStationParameters(sp);
			dataChanged = true;
		}

		VoluntaryServiceParameters vsp = getVoluntaryServiceParameters();
		if (vsp == null) {
			vsp = new VoluntaryServiceParameters(StringUtils.left(getName() + " Voluntary Service", 120));
			setVoluntaryServiceParameters(vsp);
			dataChanged = true;
		}

		List<DonGenPostFund> funds = getDonGenPostFunds();
		if (funds.isEmpty()) {
			DonGenPostFund f = new DonGenPostFund("None");
			f.setVtkInstitution(this);
			funds.add(f);
			dataChanged = true;
		}

		List<DonationReference> donationReferences = getDonationReferences();
		if (donationReferences.isEmpty()) {
			DonationReference dr = new DonationReference("None");
			dr.setVtkInstitution(this);
			donationReferences.add(dr);
			dataChanged = true;
		}
		return dataChanged;
	}

	// ---------------------------------------- Fields

	/*
	 * Only contains one item, lazy-loaded; mapping like this instead
	 * of @OneToOne so it can be optional and still lazy-load - CPB
	 */
	private List<StationParameters> stationParametersList;

	/*
	 * Only contains one item, lazy-loaded; mapping like this instead
	 * of @OneToOne so it can be optional and still lazy-load - CPB
	 */
	private List<VoluntaryServiceParameters> voluntaryServiceParametersList;

	private List<DonGenPostFund> donGenPostFunds;

	private List<DonationReference> donationReferences;

	// ---------------------------------------- Constructors

	public VtkInstitution() {
	}

	private VtkInstitution(VAFacility facility, VtkState state, VtkInstitutionType type, VtkInstitution parent) {
		super(facility, state, type, parent);
	}

	// ---------------------------------------- Business Methods

	@Transient
	public String getSta3n() {
		return StationsUtil.getThreeDigitStationNumber(getStationNumber());
	}

	@Transient
	@JsonIgnore
	public StationParameters getStationParameters() {
		List<StationParameters> l = getStationParametersList();
		return l.isEmpty() ? null : l.get(0);
	}

	public void setStationParameters(StationParameters stationParameters) {
		getStationParametersList().clear();
		if (stationParameters != null) {
			getStationParametersList().add(stationParameters);
			stationParameters.setVtkInstitution(this);
		}
	}

	@Transient
	@JsonIgnore
	public VoluntaryServiceParameters getVoluntaryServiceParameters() {
		List<VoluntaryServiceParameters> l = getVoluntaryServiceParametersList();
		return l.isEmpty() ? null : l.get(0);
	}

	public void setVoluntaryServiceParameters(VoluntaryServiceParameters vsp) {
		getVoluntaryServiceParametersList().clear();
		if (vsp != null) {
			getVoluntaryServiceParametersList().add(vsp);
			vsp.setVtkInstitution(this);
		}
	}

	// ---------------------------------------- Common Methods

	@Override
	protected boolean requiredEquals(VtkInstitution oo) {
		return new EqualsBuilder().append(getStationNumber(), oo.getStationNumber()).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(getStationNumber()).toHashCode();
	}

	// ---------------------------------------- Accessor Methods

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "vtkInstitution")
	@BatchSize(size = 500)
	@JsonIgnore
	private List<StationParameters> getStationParametersList() {
		if (stationParametersList == null)
			stationParametersList = new ArrayList<>();
		return stationParametersList;
	}

	@SuppressWarnings("unused")
	private void setStationParametersList(List<StationParameters> stationParameters) {
		this.stationParametersList = stationParameters;
	}

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "vtkInstitution")
	@BatchSize(size = 500)
	private List<VoluntaryServiceParameters> getVoluntaryServiceParametersList() {
		if (voluntaryServiceParametersList == null)
			voluntaryServiceParametersList = new ArrayList<>();
		return voluntaryServiceParametersList;
	}

	@SuppressWarnings("unused")
	private void setVoluntaryServiceParametersList(List<VoluntaryServiceParameters> voluntaryServiceParametersList) {
		this.voluntaryServiceParametersList = voluntaryServiceParametersList;
	}

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "vtkInstitution")
	@BatchSize(size = 500)
	@JsonIgnore
	public List<DonGenPostFund> getDonGenPostFunds() {
		if (donGenPostFunds == null)
			donGenPostFunds = new ArrayList<>();
		return donGenPostFunds;
	}

	public void setDonGenPostFunds(List<DonGenPostFund> donGenPostFunds) {
		this.donGenPostFunds = donGenPostFunds;
	}

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "vtkInstitution")
	@BatchSize(size = 500)
	@JsonIgnore
	public List<DonationReference> getDonationReferences() {
		if (donationReferences == null)
			donationReferences = new ArrayList<>();
		return donationReferences;
	}

	public void setDonationReferences(List<DonationReference> donationReferences) {
		this.donationReferences = donationReferences;
	}

}
