package gov.va.vss.model.facility;

import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.NONE;
import static com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility.PUBLIC_ONLY;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.DiscriminatorValue;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Transient;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.annotations.BatchSize;
import org.hibernate.annotations.Cache;
import org.hibernate.annotations.CacheConcurrencyStrategy;
import org.hibernate.validator.constraints.NotBlank;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.lookup.LookupType;
import gov.va.shared.model.lookup.sds.State;
import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.shared.util.PersistenceUtil;
import gov.va.shared.util.StationsUtil;
import gov.va.vss.model.DonGenPostFund;
import gov.va.vss.model.DonationReference;
import gov.va.vss.model.StationParameters;
import gov.va.vss.model.VoluntaryServiceParameters;
import gov.va.vss.model.lookup.FacilityType;

@Entity
@DiscriminatorValue("F")
@Cache(usage = CacheConcurrencyStrategy.NONSTRICT_READ_WRITE)
@JsonAutoDetect(fieldVisibility = NONE, getterVisibility = PUBLIC_ONLY, isGetterVisibility = PUBLIC_ONLY)
public class Facility extends AbstractUpdateableLocation<Facility> {
	private static final long serialVersionUID = -8678395783438462990L;

	public static enum FacilityValue implements LookupType {
		CENTRAL_OFFICE(2);

		private long id;

		private FacilityValue(long id) {
			this.id = id;
		}

		@Override
		public long getId() {
			return id;
		}
	}

	public static class FacilityView {
		public interface Basic {
		}

		public interface Extended extends Basic {
		}

		public interface ShowHierarchy extends Basic, BasicLocationView.ShowHierarchy {
		}
	}

	// ---------------------------------------- Fields

	@NotBlank(message = "Please enter a station number.")
	private String stationNumber;
	private FacilityType type;
	private Boolean mealPrintOCX;

	private VAFacility vaFacility;

	/**
	 * Factory method to create a new Facility which reproduces all
	 * trigger-based logic in the DB (creating related entries) - CPB
	 * 
	 * @param facility
	 * @param state
	 * @param type
	 * @param parent
	 * @return
	 */
	public static Facility createNew(VAFacility facility, State state, FacilityType type, Facility parent) {
		Facility newInstitution = new Facility(facility, state, type, null);
		newInstitution.populateMissingRequiredFields();
		return newInstitution;
	}

	public boolean populateMissingRequiredFields() {
		boolean dataChanged = false;

		StationParameters sp = getStationParameters();
		if (sp == null) {
			sp = new StationParameters(this);
			setStationParameters(sp);
			dataChanged = true;
		}

		VoluntaryServiceParameters vsp = getVoluntaryServiceParameters();
		if (vsp == null) {
			vsp = new VoluntaryServiceParameters(StringUtils.left(getName() + " Voluntary Service", 120));
			setVoluntaryServiceParameters(vsp);
			dataChanged = true;
		}

		List<DonGenPostFund> funds = getDonGenPostFunds();
		if (funds.isEmpty()) {
			DonGenPostFund f = new DonGenPostFund("None");
			f.setFacility(this);
			funds.add(f);
			dataChanged = true;
		}

		List<DonationReference> donationReferences = getDonationReferences();
		if (donationReferences.isEmpty()) {
			DonationReference dr = new DonationReference("None");
			dr.setFacility(this);
			donationReferences.add(dr);
			dataChanged = true;
		}
		return dataChanged;
	}

	// ---------------------------------------- Fields

	/*
	 * Only contains one item, lazy-loaded; mapping like this instead
	 * of @OneToOne so it can be optional and still lazy-load - CPB
	 */
	private List<StationParameters> stationParametersList;

	/*
	 * Only contains one item, lazy-loaded; mapping like this instead
	 * of @OneToOne so it can be optional and still lazy-load - CPB
	 */
	private List<VoluntaryServiceParameters> voluntaryServiceParametersList;

	private List<DonGenPostFund> donGenPostFunds;

	private List<DonationReference> donationReferences;

	// ---------------------------------------- Constructors

	public Facility() {
	}

	private Facility(VAFacility facility, State state, FacilityType type, Facility parent) {
		super(facility.getName(), facility.getAddressLine1(), facility.getAddressLine2(), facility.getCity(), state,
				facility.getZip(), parent);
		this.stationNumber = facility.getStationNumber();
		this.type = type;
		this.vaFacility = facility;
	}

	// ---------------------------------------- Business Methods

	@Override
	@Transient
	@JsonView(FacilityView.Extended.class)
	public String getScale() {
		return "Facility";
	}

	@Transient
	public String getDisplayName() {
		VAFacility facility = getVaFacility();
		if (facility == null || StringUtils.isNotBlank(getName()))
			return getName() + " (" + getStationNumber() + ")";
		return getVaFacility().getDisplayName();
	}

	@Transient
	public String getDisplayNameAbbreviated() {
		return StringUtils.abbreviateMiddle(getDisplayName(), "..", 50);
	}

	@Transient
	public boolean isCentralOffice() {
		return FacilityValue.CENTRAL_OFFICE.getId() == getId();
	}

	@Transient
	@JsonView(FacilityView.Extended.class)
	public boolean isLinkedToVAFacility() {
		return getVaFacility() != null;
	}

	@Transient
	public String getSta3n() {
		return StationsUtil.getThreeDigitStationNumber(getStationNumber());
	}

	@Transient
	@JsonIgnore
	public StationParameters getStationParameters() {
		List<StationParameters> l = getStationParametersList();
		return l.isEmpty() ? null : l.get(0);
	}

	public void setStationParameters(StationParameters stationParameters) {
		getStationParametersList().clear();
		if (stationParameters != null) {
			getStationParametersList().add(stationParameters);
			stationParameters.setFacility(this);
		}
	}

	@Transient
	@JsonIgnore
	public VoluntaryServiceParameters getVoluntaryServiceParameters() {
		List<VoluntaryServiceParameters> l = getVoluntaryServiceParametersList();
		return l.isEmpty() ? null : l.get(0);
	}

	public void setVoluntaryServiceParameters(VoluntaryServiceParameters vsp) {
		getVoluntaryServiceParametersList().clear();
		if (vsp != null) {
			getVoluntaryServiceParametersList().add(vsp);
			vsp.setFacility(this);
		}
	}

	// ---------------------------------------- Common Methods

	@Override
	protected boolean requiredEquals(AbstractLocation oo) {
		/*
		 * Safe cast due to AbstractBusinessKeyOwner logic; since we need a
		 * field in the concrete class, we need to ensure we aren't trying to
		 * cast a proxy, which would fail - CPB
		 */
		Facility f = (Facility) PersistenceUtil.initializeAndUnproxy(oo);
		return new EqualsBuilder().append(getStationNumber(), f.getStationNumber()).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(getStationNumber()).toHashCode();
	}

	// ---------------------------------------- Accessor Methods

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "facility")
	@BatchSize(size = 500)
	@JsonIgnore
	private List<StationParameters> getStationParametersList() {
		if (stationParametersList == null)
			stationParametersList = new ArrayList<>();
		return stationParametersList;
	}

	@SuppressWarnings("unused")
	private void setStationParametersList(List<StationParameters> stationParameters) {
		this.stationParametersList = stationParameters;
	}

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "facility")
	@BatchSize(size = 500)
	private List<VoluntaryServiceParameters> getVoluntaryServiceParametersList() {
		if (voluntaryServiceParametersList == null)
			voluntaryServiceParametersList = new ArrayList<>();
		return voluntaryServiceParametersList;
	}

	@SuppressWarnings("unused")
	private void setVoluntaryServiceParametersList(List<VoluntaryServiceParameters> voluntaryServiceParametersList) {
		this.voluntaryServiceParametersList = voluntaryServiceParametersList;
	}

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "facility")
	@BatchSize(size = 500)
	@JsonIgnore
	public List<DonGenPostFund> getDonGenPostFunds() {
		if (donGenPostFunds == null)
			donGenPostFunds = new ArrayList<>();
		return donGenPostFunds;
	}

	public void setDonGenPostFunds(List<DonGenPostFund> donGenPostFunds) {
		this.donGenPostFunds = donGenPostFunds;
	}

	@OneToMany(cascade = CascadeType.ALL, orphanRemoval = true, fetch = FetchType.LAZY, mappedBy = "facility")
	@BatchSize(size = 500)
	@JsonIgnore
	public List<DonationReference> getDonationReferences() {
		if (donationReferences == null)
			donationReferences = new ArrayList<>();
		return donationReferences;
	}

	public void setDonationReferences(List<DonationReference> donationReferences) {
		this.donationReferences = donationReferences;
	}

	@Column(length = 7, nullable = false)
	@JsonView(FacilityView.Basic.class)
	public String getStationNumber() {
		return stationNumber;
	}

	public void setStationNumber(String stationNumber) {
		this.stationNumber = stationNumber;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "TypeFK", nullable = false)
	@BatchSize(size = 500)
	@JsonView(FacilityView.Extended.class)
	public FacilityType getType() {
		return type;
	}

	public void setType(FacilityType type) {
		this.type = type;
	}

	@JsonIgnore
	public Boolean getMealPrintOCX() {
		return mealPrintOCX;
	}

	public void setMealPrintOCX(Boolean mealPrintOCX) {
		this.mealPrintOCX = mealPrintOCX;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "SDS_INSTITUTION_ID")
	@BatchSize(size = 500)
	@JsonIgnore
	public VAFacility getVaFacility() {
		return vaFacility;
	}

	public void setVaFacility(VAFacility facility) {
		this.vaFacility = facility;
	}

}
