package gov.va.vss.model.jobExecution;

import java.time.ZonedDateTime;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.persistence.AttributeOverride;
import javax.persistence.AttributeOverrides;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.FetchType;
import javax.persistence.JoinColumn;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.Table;
import javax.persistence.Transient;
import javax.validation.constraints.NotNull;

import org.apache.commons.lang3.builder.CompareToBuilder;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.hibernate.annotations.BatchSize;
import org.hibernate.annotations.SortNatural;
import org.springframework.batch.core.JobExecution;

import com.fasterxml.jackson.annotation.JsonProperty;
import com.google.common.base.Predicate;
import com.google.common.collect.Sets;

import gov.va.shared.model.AppUser;
import gov.va.shared.model.core.AbstractIdentifiedPersistent;
import gov.va.shared.model.lookup.sds.VAFacility;
import gov.va.vss.model.lookup.JobExecutionType;
import gov.va.vss.util.DateUtil;

/**
 * Generic class representing the result of some job (whether it be a spring
 * batch job, manual Java code or JMS endpoint)
 * 
 * @author vhaisdbarryc
 */
@Entity
@Table(name = "JOB_EXECUTION_RESULT", schema = "CORE")
@AttributeOverrides({ @AttributeOverride(name = "id", column = @Column(name = "JOB_EXECUTION_RESULT_ID") ) })
public class JobExecutionResult extends AbstractIdentifiedPersistent<JobExecutionResult>
		implements Comparable<JobExecutionResult> {
	private static final long serialVersionUID = 4416728108665979313L;

	// --------------------------------------- Fields

	@NotNull
	private JobExecutionType type;

	@NotNull
	private ZonedDateTime startTime;
	private ZonedDateTime endTime;
	private AppUser initiator;

	private Long referenceJobExecutionId;
	private VAFacility referenceInstitution;

	private SortedSet<JobExecutionResultItem> jobExecutionResultItems;

	// --------------------------------------- Constructors

	public JobExecutionResult() {
	}

	public JobExecutionResult(JobExecution jobExecution, JobExecutionType type, AppUser initiator) {
		this(type, initiator, DateUtil.getZonedDateTime(jobExecution.getStartTime()));
		setReferenceJobExecutionId(jobExecution.getId());
		setEndTime(DateUtil.getZonedDateTime(jobExecution.getEndTime()));
	}

	public JobExecutionResult(JobExecutionType type, AppUser initiator, ZonedDateTime startTime) {
		setType(type);
		setInitiator(initiator);
		setStartTime(startTime);
	}

	// --------------------------------------- Business Methods

	@Transient
	public SortedSet<JobExecutionResultItem> getResultItemsByCriteria(final Boolean successful) {
		SortedSet<JobExecutionResultItem> assignments = getJobExecutionResultItems();
		if (assignments.isEmpty())
			return assignments;

		Predicate<JobExecutionResultItem> predicate = new Predicate<JobExecutionResultItem>() {
			@Override
			public boolean apply(JobExecutionResultItem o) {
				if (successful != null && !successful.equals(o.isItemSuccessful()))
					return false;

				return true;
			}
		};
		SortedSet<JobExecutionResultItem> filteredRelationships = Sets.filter(assignments, predicate);
		return filteredRelationships;
	}

	@Transient
	public int getFailedResultItemCount() {
		return getResultItemsByCriteria(false).size();
	}

	// --------------------------------------- Common Methods

	@Override
	public int compareTo(JobExecutionResult u) {
		if (equals(u))
			return 0;

		return new CompareToBuilder().append(getStartTime(), u.getStartTime()).append(getEndTime(), u.getEndTime())
				.append(nullSafeGetId(getType()), nullSafeGetId(u.getType())).toComparison() > 0 ? 1 : -1;
	}

	@Override
	protected boolean requiredEquals(JobExecutionResult u) {
		return new EqualsBuilder().append(getStartTime(), u.getStartTime()).append(getEndTime(), u.getEndTime())
				.append(nullSafeGetId(getType()), nullSafeGetId(u.getType())).isEquals();
	}

	@Override
	protected int requiredHashCode() {
		return new HashCodeBuilder().append(getStartTime()).append(getEndTime()).append(nullSafeGetId(getType()))
				.toHashCode();
	}

	// --------------------------------------- Accessor Methods

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "STD_JOB_EXECUTION_TYPE_ID")
	public JobExecutionType getType() {
		return type;
	}

	public void setType(JobExecutionType type) {
		this.type = type;
	}

	@OneToMany(mappedBy = "jobExecutionResult", fetch = FetchType.LAZY, cascade = CascadeType.ALL)
	@SortNatural
	@BatchSize(size = 20)
	@JsonProperty
	public SortedSet<JobExecutionResultItem> getJobExecutionResultItems() {
		if (jobExecutionResultItems == null)
			jobExecutionResultItems = new TreeSet<>();
		return jobExecutionResultItems;
	}

	public void setJobExecutionResultItems(SortedSet<JobExecutionResultItem> jobExecutionResultItems) {
		this.jobExecutionResultItems = jobExecutionResultItems;
	}

	@Column(name = "START_TIME")
	public ZonedDateTime getStartTime() {
		return startTime;
	}

	public void setStartTime(ZonedDateTime startTime) {
		this.startTime = startTime;
	}

	@Column(name = "JOB_EXECUTION_ID")
	public Long getReferenceJobExecutionId() {
		return referenceJobExecutionId;
	}

	public void setReferenceJobExecutionId(Long referenceJobExecutionId) {
		this.referenceJobExecutionId = referenceJobExecutionId;
	}

	@Column(name = "END_TIME")
	public ZonedDateTime getEndTime() {
		return endTime;
	}

	public void setEndTime(ZonedDateTime endTime) {
		this.endTime = endTime;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "VA_INSTITUTION_ID")
	public VAFacility getReferenceInstitution() {
		return referenceInstitution;
	}

	public void setReferenceInstitution(VAFacility referenceInstitution) {
		this.referenceInstitution = referenceInstitution;
	}

	@ManyToOne(fetch = FetchType.LAZY)
	@JoinColumn(name = "EXECUTE_REQUEST_BY")
	public AppUser getInitiator() {
		return initiator;
	}

	public void setInitiator(AppUser initiator) {
		this.initiator = initiator;
	}

}
