package gov.va.vss.service.messaging.listeners;

import java.io.IOException;
import java.time.ZonedDateTime;
import java.util.Map;

import javax.annotation.Resource;
import javax.jms.JMSException;
import javax.jms.MapMessage;
import javax.jms.Message;
import javax.jms.Session;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jms.core.JmsTemplate;
import org.springframework.jms.core.MessageCreator;
import org.springframework.stereotype.Service;

import gov.va.shared.model.lookup.LookupType;
import gov.va.shared.util.LookupUtil;
import gov.va.vss.model.VistaStationData;
import gov.va.vss.model.jobExecution.JobExecutionResult;
import gov.va.vss.persistence.dao.VistaStationDataDAO;
import gov.va.vss.persistence.dao.jobExecution.JobExecutionResultDAO;
import gov.va.vss.service.batch.JobExecutionResultItemService;
import gov.va.vss.util.SerializationUtil;

@Service
public class CommonEventsJMSService {
	private static final Logger log = LoggerFactory.getLogger(CommonEventsJMSService.class);

	private static final String PARAM_COMMON_EVENTS_TYPE = "commonEventsType";
	private static final String PARAM_JOB_EXECUTION_RESULT_ID = "jobExecutionResultId";
	private static final String PARAM_EVENT = "event";
	private static final String PARAM_STATION_NUMBER = "stationNumber";

	@Resource(name = "jmsTemplateCommonEvents")
	private JmsTemplate commonEventsJmsTemplate;

	@Autowired
	private JobExecutionResultDAO jobExecutionResultDAO;
	@Autowired
	private JobExecutionResultItemService jobExecutionResultItemService;
	@Autowired
	private VistaStationDataDAO vistaStationDataDAO;

	/**
	 * Method that gets called to send JMS message requesting a Vista update
	 * 
	 * @throws IOException
	 */
	public void requestProcessSomething(final long jobExecutionResultId, final String stationNumber)
			throws IOException {
		final byte[] serializedEvent = SerializationUtil.serializeObject("some data");

		commonEventsJmsTemplate.send(new MessageCreator() {
			@Override
			public Message createMessage(Session session) throws JMSException {
				MapMessage message = session.createMapMessage();
				message.setLong(PARAM_COMMON_EVENTS_TYPE, CommonEventsType.PROCESS_SOME_EVENT.getId());
				message.setLong(PARAM_JOB_EXECUTION_RESULT_ID, jobExecutionResultId);
				message.setString(PARAM_STATION_NUMBER, stationNumber);
				message.setObject(PARAM_EVENT, serializedEvent);
				return message;
			}
		});
	}

	/**
	 * Method that gets activated when the JMS message generated by
	 * requestVistaUpdate() is received.
	 * 
	 * @throws IOException
	 * @throws ClassNotFoundException
	 */
	public void handleMessage(Map<String, Object> messageData) throws ClassNotFoundException, IOException {
		CommonEventsType type = LookupUtil.getById(CommonEventsType.class,
				(Long) messageData.get(PARAM_COMMON_EVENTS_TYPE));

		switch (type) {
		case PROCESS_SOME_EVENT:
			String stationNumber = (String) messageData.get(PARAM_STATION_NUMBER);

			/*
			 * May be null if the JobExecutionResult has been purged since the
			 * original message was sent - CPB
			 */
			JobExecutionResult jobExecutionResult = jobExecutionResultDAO
					.findByPrimaryKey((long) messageData.get(PARAM_JOB_EXECUTION_RESULT_ID));
			byte[] eventBytes = (byte[]) messageData.get(PARAM_EVENT);
			String event = (String) SerializationUtil.deserializeObject(eventBytes);
			ZonedDateTime start = ZonedDateTime.now();

			try {
				VistaStationData vsd = vistaStationDataDAO.findByStationNumber(stationNumber);
				// patientAutoInactivationService.processPatientEventFlaggedByVista(event,
				// jobExecutionResult, vsd);
			} catch (Exception e) {
				String msg = "Encountered a " + e.getClass().getSimpleName() + " when processing the "
						+ String.class.getSimpleName() + " with value " + event;
				log.error(msg + " (set log to DEBUG to see full stack trace - continuing with other events...)");
				log.debug("Full error: ", e);

				e.printStackTrace();

				if (jobExecutionResult != null)
					jobExecutionResultItemService.saveItemResultInNewTransaction(jobExecutionResult, start,
							ZonedDateTime.now(), false, msg, e);

				throw new RuntimeException(e);
			}
			break;
		default:
			throw new AssertionError("Unexpected value of CommonEventsType " + type);
		}
	}

	public static enum CommonEventsType implements LookupType {
		PROCESS_SOME_EVENT(1);

		private long id;

		private CommonEventsType(long id) {
			this.id = id;
		}

		public long getId() {
			return id;
		}

	}

}
