package gov.va.vss.web.donation;

import java.util.LinkedHashMap;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.servlet.http.HttpServletRequest;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.ModelMap;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.SessionAttributes;
import org.springframework.web.bind.support.SessionStatus;

import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.Permission;
import gov.va.shared.model.Permission.PermissionType;
import gov.va.shared.model.lookup.sds.State;
import gov.va.shared.util.StringUtil;
import gov.va.vss.model.DonationSummary;
import gov.va.vss.model.Donor;
import gov.va.vss.model.Donor.DonorView;
import gov.va.vss.model.lookup.VtkDonorType;
import gov.va.vss.service.validation.ServiceValidationException;
import gov.va.vss.web.AbstractAppController;
import gov.va.vss.web.ValidationException;
import gov.va.vss.web.breadcrumbs.Breadcrumb;

@Controller
@SessionAttributes(value = { AbstractAppController.DEFAULT_COMMAND_NAME })
public class DonorController extends AbstractAppController {

	@Autowired
	private DonorValidator donorValidator;

	@RequestMapping("/donorCreate.htm")
	@Breadcrumb("Create Donor")
	@PreAuthorize("hasAuthority('" + Permission.DONATION_CREATE + "')")

	public String createDonor(ModelMap model, HttpServletRequest request) {
		Donor donor = new Donor();

		DonorCommand command = new DonorCommand(donor);

		model.addAttribute(DEFAULT_COMMAND_NAME, command);

		createReferenceData(model);

		return "createDonor";
	}

	@RequestMapping("/donorEdit.htm")
	@Breadcrumb("Edit Donor")
	@PreAuthorize("hasAuthority('" + Permission.DONATION_READ + "')")
	public String editDonor(@RequestParam long id, ModelMap model, HttpServletRequest request) {
		Donor donor = donorDAO.findRequiredByPrimaryKey(id);

		DonorCommand command = new DonorCommand(donor);

		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		createReferenceData(model);
		updateModelForPermissionsIfNecessary(model, PermissionType.DONATION_CREATE);

		LinkedHashMap<DonationSummary, Boolean> results = new LinkedHashMap<>();
		for (DonationSummary s : donor.getDonations()) {
			results.put(s, dateUtil.isDonationSummaryEditable(s.getDonationDate()));
		}
		model.addAttribute("donationMap", results);

		return "editDonor";
	}

	private void createReferenceData(ModelMap model) {
	}

	@RequestMapping("/donorSubmit.htm")
	@PreAuthorize("hasAuthority('" + Permission.DONATION_CREATE + "')")
	public String submitDonor(@ModelAttribute(DEFAULT_COMMAND_NAME) DonorCommand command, BindingResult result,
			SessionStatus status, ModelMap model, HttpServletRequest request) throws ValidationException {
		Donor donor = command.getDonor();

		boolean isEdit = donor.isPersistent();

		if (isEdit) {
			if ("individual".equals(command.getDesiredIndividualType()) && donor.getVolunteer() != null) {
				donor.setVolunteer(null);
			}
		}

		/* Validation step (JSR303, other custom logic in the validator) */
		donorValidator.validate(command, result, false, "donor");
		boolean hasErrors = result.hasErrors();

		if (!hasErrors) {
			try {
				donor = donorService.saveOrUpdate(donor);
				userNotifier.notifyUserOnceWithMessage(request,
						getMessage(isEdit ? "donor.update.success" : "donor.create.success"));
			} catch (ServiceValidationException e) {
				webValidationService.handle(e, result);
				hasErrors = true;
			}
		}

		if (hasErrors) {
			createReferenceData(model);
			return isEdit ? "editDonor" : "createDonor";
		} else {
			status.setComplete();
			return "redirect:/donorEdit.htm?id=" + donor.getId();
		}
	}

	@RequestMapping("/findDonors")
	@PreAuthorize("hasAuthority('" + Permission.DONATION_READ + "')")
	@JsonView(DonorView.Basic.class)
	public @ResponseBody SortedSet<Donor> findDonors(@RequestParam VtkDonorType donorType,
			@RequestParam(required = false) String name, @RequestParam(required = false) String city,
			@RequestParam(required = false) State state, @RequestParam(required = false) String zip,
			@RequestParam(required = false) String email, @RequestParam(required = false) String phone) {
		if (StringUtil.allBlank(name, city, state != null ? state.getId().toString() : "", zip, email, phone))
			throw new IllegalArgumentException("Please specify at least one piece of search criteria");

		String firstName = null;
		String lastName = null;
		String middleName = null;
		if (StringUtils.isNotBlank(name)) {
			String[] nameComponents = StringUtil.parseNameComponents(name);
			lastName = nameComponents[0];
			firstName = nameComponents[1];
			middleName = nameComponents[2];
		}

		SortedSet<Donor> donors = new TreeSet<>(
				donorDAO.findByCriteria(donorType, firstName, middleName, lastName, city, state, zip, email, phone));

		return donors;
	}

	@RequestMapping("/donor/donorLinkVolunteer")
	@PreAuthorize("hasAuthority('" + Permission.DONATION_CREATE + "')")
	public @ResponseBody Donor linkVolunteer(@RequestParam(required = false) Long donorId,
			@RequestParam Long volunteerId) throws ValidationException {
		return donorService.linkVolunteer(donorId, volunteerId);
	}

	@RequestMapping("/donor/donorLinkOrganization")
	@PreAuthorize("hasAuthority('" + Permission.DONATION_CREATE + "')")
	public @ResponseBody Donor linkOrganization(@RequestParam(required = false) Long donorId, @RequestParam Long orgId)
			throws ValidationException {

		return donorService.linkOrganization(donorId, orgId);
	}

}
