package gov.va.vss.web.facility;

import static gov.va.vss.model.lookup.FacilityType.FacilityTypeValue.GAMES;
import static gov.va.vss.model.lookup.FacilityType.FacilityTypeValue.TIMEKEEPING;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.servlet.http.HttpServletRequest;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.dao.DataIntegrityViolationException;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.ModelMap;
import org.springframework.validation.BindingResult;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.SessionAttributes;
import org.springframework.web.bind.support.SessionStatus;

import com.fasterxml.jackson.annotation.JsonView;

import gov.va.shared.model.Permission;
import gov.va.shared.model.Permission.PermissionType;
import gov.va.shared.model.lookup.sds.State;
import gov.va.shared.persistence.lookup.sds.VAFacilityDAO.QuickSearchResult;
import gov.va.shared.util.SecurityUtil;
import gov.va.shared.util.WebUtil;
import gov.va.vss.model.BenefitingServiceRole;
import gov.va.vss.model.facility.Facility;
import gov.va.vss.model.facility.FacilityNode;
import gov.va.vss.model.facility.Location;
import gov.va.vss.model.facility.VtkAdministrativeUnit;
import gov.va.vss.model.facility.VtkAdministrativeUnit.VtkAdministrativeUnitView;
import gov.va.vss.model.lookup.FacilityType.FacilityTypeValue;
import gov.va.vss.service.validation.ServiceValidationException;
import gov.va.vss.web.AbstractAppController;
import gov.va.vss.web.ValidationException;
import gov.va.vss.web.breadcrumbs.Breadcrumb;

@Controller
@SessionAttributes(value = { AbstractAppController.DEFAULT_COMMAND_NAME })
public class FacilityController extends AbstractAppController {

	@Autowired
	private FacilityValidator facilityValidator;

	@RequestMapping("/facilityCreate.htm")
	@Breadcrumb("Create Facility")
	@PreAuthorize("hasAnyAuthority('" + Permission.FACILITY_CREATE + "')")
	public String createFacility(ModelMap model, HttpServletRequest request) {
		Facility facility = new Facility();
		FacilityCommand command = new FacilityCommand(facility);
		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		createReferenceData(model);
		return "facilityCreate";
	}

	@RequestMapping("/facilityEdit.htm")
	@Breadcrumb("Edit Facilities")
	@PreAuthorize("hasAnyAuthority('" + Permission.FACILITY_EDIT_ALL + ", " + Permission.FACILITY_EDIT_CURRENT + "')")
	public String editFacility(@RequestParam(required = false) Long id, ModelMap model, HttpServletRequest request) {
		FacilityCommand command = new FacilityCommand();
		if (SecurityUtil.hasAllPermissionsAtCurrentFacility(PermissionType.FACILITY_EDIT_ALL)) {
			if (id != null) {
				Facility facility = facilityDAO.findRequiredByPrimaryKey(id);
				command = new FacilityCommand(facility);
			}
		} else if (SecurityUtil.hasAllPermissionsAtCurrentFacility(PermissionType.FACILITY_EDIT_CURRENT)) {
			command = new FacilityCommand(getFacilityContext());
		} else {
			throw new SecurityException();
		}

		model.addAttribute(DEFAULT_COMMAND_NAME, command);
		createReferenceData(model);
		return "facilityEdit";
	}

	private void createReferenceData(ModelMap model) {
		model.put("allTypes", facilityTypeDAO.findByLookups(TIMEKEEPING, GAMES).values());
		model.put("allVISNs", vtkAdministrativeUnitDAO.findAllSorted());
		model.put("allFacilities", facilityDAO.findAllSorted());
		
		WebUtil.addEnumToModel(FacilityTypeValue.class, model);
	}

	@RequestMapping("/facilitySubmit.htm")
	@PreAuthorize("hasAnyAuthority('" + Permission.FACILITY_CREATE + ", " + Permission.FACILITY_EDIT_CURRENT + ", "
			+ Permission.FACILITY_EDIT_ALL + "')")
	public String submitFacility(@ModelAttribute(DEFAULT_COMMAND_NAME) FacilityCommand command, BindingResult result,
			SessionStatus status, ModelMap model, HttpServletRequest request) throws ValidationException {
		Facility facility = command.getFacility();
		boolean isEdit = facility.isPersistent();

		Facility parent = command.getParentId() == null ? null
				: facilityDAO.findRequiredByPrimaryKey(command.getParentId());
		facility.setParent(parent);

		/* Validation step (JSR303, other custom logic in the validator) */
		facilityValidator.validate(command, result, false, "facility");
		boolean hasErrors = result.hasErrors();

		if (!hasErrors) {
			try {
				facility = facilityService.saveOrUpdate(facility);
				userNotifier.notifyUserOnceWithMessage(request,
						getMessage(isEdit ? "facility.update.success" : "facility.create.success"));
			} catch (DataIntegrityViolationException e) {
				hasErrors = true;
				webValidationService.handle(e, result, "facility.create.error.dataIntegrityError");
			} catch (ServiceValidationException e) {
				webValidationService.handle(e, result);
				hasErrors = true;
			}
		}

		if (hasErrors) {
			createReferenceData(model);
			return isEdit ? "facilityEdit" : "facilityCreate";
		} else {
			status.setComplete();
			return "redirect:/facilityEdit.htm?id=" + facility.getId();
		}
	}

	// ======================= Locations ==============================

	// @RequestMapping("/locationList")
	// @PreAuthorize("hasAnyAuthority('" + Permission.ORG_CODE_NATIONAL_READ +
	// ", " + Permission.ORG_CODE_LOCAL_READ
	// + "')")
	// public @ResponseBody List<NationalOfficial>
	// getOfficialForFacility(@RequestParam long facilityId) {
	// List<NationalOfficial> officials =
	// nationalOfficialDAO.findByCriteria(orgId, null);
	// return officials;
	// }
	//
	// @RequestMapping("/nationalOfficalCreateOrUpdate")
	// @PreAuthorize("hasAnyAuthority('" + Permission.ORG_CODE_NATIONAL_CREATE +
	// ", " + Permission.ORG_CODE_LOCAL_CREATE
	// + "')")
	// public @ResponseBody NationalOfficial
	// createOrUpdateNationalOfficial(@RequestParam Long facilityId,
	// @RequestParam Long nationalOfficialId, @RequestParam String lastName,
	// @RequestParam String firstName,
	// @RequestParam String middleName, @RequestParam String suffix,
	// @RequestParam String prefix,
	// @RequestParam String title, @RequestParam boolean certifyingOfficial,
	// @RequestParam String email,
	// @RequestParam String streetAddress, @RequestParam String city,
	// @RequestParam State state,
	// @RequestParam String zip, @RequestParam String phone, @RequestParam
	// StdVAVSTitle stdVAVSTitle,
	// @DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam
	// LocalDate vavsStartDate,
	// @DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam
	// LocalDate vavsEndDate,
	// @RequestParam boolean nationalCommitteeMember,
	// @DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam
	// LocalDate nacStartDate,
	// @DateTimeFormat(pattern = DateUtil.TWO_DIGIT_DATE_ONLY) @RequestParam
	// LocalDate nacEndDate)
	// throws ServiceValidationException {
	//
	// if (facilityId == null) {
	// throw new IllegalArgumentException("Facility Id cannot be null");
	// }
	//
	// NationalOfficial official = null;
	// if (nationalOfficialId == null) {
	// official = new NationalOfficial();
	// AbstractBasicFacility<?> org = facilityDAO.findByPrimaryKey(facilityId);
	// if (org == null) {
	// throw new IllegalArgumentException("Facility not found for id " +
	// facilityId);
	// }
	// official.setFacility(org);
	// } else {
	// official =
	// nationalOfficialDAO.findRequiredByPrimaryKey(nationalOfficialId);
	// }
	//
	// official.setLastName(lastName);
	// official.setFirstName(firstName);
	// official.setMiddleName(middleName);
	// official.setSuffix(suffix);
	// official.setPrefix(prefix);
	// official.setTitle(title);
	// official.setCertifyingOfficial(certifyingOfficial);
	// official.setEmail(email);
	// official.setStreetAddress(streetAddress);
	// official.setCity(city);
	// official.setZip(zip);
	// official.setPhone(phone);
	// official.setStdVAVSTitle(stdVAVSTitle);
	// official.setVavsStartDate(vavsStartDate);
	// official.setVavsEndDate(vavsEndDate);
	// official.setNationalCommitteeMember(nationalCommitteeMember);
	// official.setNacStartDate(nacStartDate);
	// official.setNacEndDate(nacEndDate);
	// official.setState(state);
	//
	// official = nationalOfficialService.saveOrUpdate(official);
	//
	// return official;
	// }
	//
	// @RequestMapping("/nationalOfficialDelete")
	// @PreAuthorize("hasAnyAuthority('" + Permission.ORG_CODE_NATIONAL_CREATE +
	// ", " + Permission.ORG_CODE_LOCAL_CREATE
	// + "')")
	// public @ResponseBody boolean deleteNationalOfficial(@RequestParam long
	// officialId)
	// throws ServiceValidationException {
	// nationalOfficialService.delete(officialId);
	// return true;
	// }
	//
	// @RequestMapping("/getFacilitySearchStations")
	// public @ResponseBody List<BasicFacility<?>>
	// getStationsForVolunteerSearch() {
	// return facilityDAO.findByCriteria(null, true, true,
	// Arrays.asList(getFacilityContext().getId()), null,
	// null, null);
	// }
	//
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@RequestMapping("/findFacilitiesForHierarchyDisplay")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	@JsonView(VtkAdministrativeUnitView.ShowHierarchy.class)
	public @ResponseBody SortedSet<? extends FacilityNode<?>> findFacilitiesForHierarchyDisplay() {
		SortedSet facilities = vtkAdministrativeUnitDAO.findAllSorted();

		SortedSet<? extends FacilityNode<?>> rootFacilities = facilityDAO.findRootFacilities();
		for (Object f : facilities)
			rootFacilities.removeAll(((VtkAdministrativeUnit) f).getFacilityChildren());

		facilities.add(new FacilityNode() {
			@Override
			public SortedSet getFacilityChildren() {
				return rootFacilities;
			}

			@Override
			public Long getId() {
				return -1L;
			}

			@Override
			public String getDisplayName() {
				return "(No VISN Assigned)";
			}

			@Override
			public int compareTo(Object o) {
				if (equals(o))
					return 0;
				return 1;
			}

			@Override
			public boolean isActive() {
				return true;
			}
		});

		return facilities;
	}

	@RequestMapping("/unlinkSDSFacilityFromFacility")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean unlinkSDSFacilityFromFacility(@RequestParam long facilityId) {
		facilityService.unlinkSDSFacilityFromFacility(facilityId);
		return true;
	}

	@RequestMapping("/linkSDSFacilityToFacility")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean linkSDSFacilityToFacility(@RequestParam long facilityId,
			@RequestParam long vaFacilityId) {
		facilityService.linkSDSFacilityToFacility(facilityId, vaFacilityId);
		return true;
	}

	@RequestMapping("/facility/location")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody SortedSet<Location> getLocationsForFacility(@RequestParam long facilityId) {
		SortedSet<Location> results = facilityDAO.findByCriteria(null, facilityId, Location.class);
		return results;
	}

	@RequestMapping("/facility/location/saveOrUpdate")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody Location locationSaveOrUpdate(@RequestParam long facilityId,
			@RequestParam(required = false) Long locationId, @RequestParam String name,
			@RequestParam String addressLine1, @RequestParam String addressLine2, @RequestParam String city,
			@RequestParam State state, @RequestParam String zip, @RequestParam String contactName,
			@RequestParam String contactRole, @RequestParam String contactPhone, @RequestParam String contactEmail,
			@RequestParam boolean active) throws ServiceValidationException {
		Location location = new Location();
		if (locationId != null) {
			location = locationDAO.findRequiredByPrimaryKey(locationId);
		} else {
			location.setParent(facilityDAO.findRequiredByPrimaryKey(facilityId));
		}
		location.setName(name);
		location.setAddressLine1(addressLine1);
		location.setAddressLine2(addressLine2);
		location.setCity(city);
		location.setState(state);
		location.setZip(zip);
		location.setContactName(contactName);
		location.setContactRole(contactRole);
		location.setContactEmail(contactEmail);
		location.setContactPhone(contactPhone);
		location.setActive(active);
		location = locationService.saveOrUpdate(location);
		return location;
	}

	@RequestMapping("/facility/location/delete")
	// @PreAuthorize("hasAuthority('" + Permission.VOLUNTEER_VIEW + "')")
	public @ResponseBody boolean locationDelete(@RequestParam long locationId) throws ServiceValidationException {
		List<BenefitingServiceRole> serviceRolesReferencingLocation = benefitingServiceRoleDAO.findByCriteria(null,
				false, true, Arrays.asList(locationId));
		if (serviceRolesReferencingLocation.isEmpty()) {
			locationService.delete(locationId);
		} else {
			Location location = locationDAO.findRequiredByPrimaryKey(locationId);
			location.setActive(false);
			location = locationService.saveOrUpdate(location);
		}
		return true;
	}

	@RequestMapping("/facility/quickSearch")
	public @ResponseBody Map<String, Object> findAppUserByNameOrUsername(@RequestParam int draw,
			@RequestParam int start, @RequestParam int length, @RequestParam(name = "search[value]") String searchValue,
			@RequestParam(name = "search[regex]") boolean searchIsRegex) {
		Map<String, Object> resultMap = new HashMap<>();

		SortedSet<QuickSearchResult> results = null;
		if (StringUtils.isNotBlank(searchValue)) {
			results = new TreeSet<>(vaFacilityDAO.findByCriteria(searchValue, length));
		} else {
			results = new TreeSet<>();
		}
		resultMap.put("data", results);

		resultMap.put("draw", draw);

		return resultMap;
	}

}
